@@ -1,11 +1,44 @@
 use super::{Error, Session};
+
+#[cfg(feature = "process-mux")]
+use super::process_impl;
+
+#[cfg(feature = "native-mux")]
+use super::native_mux_impl;
+
 use std::borrow::Cow;
+use std::ffi::OsString;
+use std::fs;
 use std::path::{Path, PathBuf};
 use std::process::Stdio;
-use tempfile::Builder;
-use tokio::io::AsyncReadExt;
+use std::str;
+
+use dirs::state_dir;
+use once_cell::sync::OnceCell;
+use tempfile::{Builder, TempDir};
 use tokio::process;
 
+/// The returned `&'static Path` can be coreced to any lifetime.
+fn get_default_control_dir<'a>() -> Result<&'a Path, Error> {
+    static DEFAULT_CONTROL_DIR: OnceCell<Option<Box<Path>>> = OnceCell::new();
+
+    DEFAULT_CONTROL_DIR
+        .get_or_try_init(|| {
+            if let Some(state_dir) = state_dir() {
+                fs::create_dir_all(&state_dir).map_err(Error::Connect)?;
+
+                Ok(Some(state_dir.into_boxed_path()))
+            } else {
+                Ok(None)
+            }
+        })
+        .map(|default_control_dir| {
+            default_control_dir
+                .as_deref()
+                .unwrap_or_else(|| Path::new("./"))
+        })
+}
+
 /// Build a [`Session`] with options.
 #[derive(Debug, Clone)]
 pub struct SessionBuilder {
@@ -17,6 +50,8 @@ pub struct SessionBuilder {
     known_hosts_check: KnownHosts,
     control_dir: Option<PathBuf>,
     config_file: Option<PathBuf>,
+    compression: Option<bool>,
+    user_known_hosts_file: Option<Box<Path>>,
 }
 
 impl Default for SessionBuilder {
@@ -30,6 +65,8 @@ impl Default for SessionBuilder {
             known_hosts_check: KnownHosts::Add,
             control_dir: None,
             config_file: None,
+            compression: None,
+            user_known_hosts_file: None,
         }
     }
 }
@@ -90,6 +127,8 @@ impl SessionBuilder {
     /// be created.
     ///
     /// If not set, `./` will be used (the current directory).
+    #[cfg(not(windows))]
+    #[cfg_attr(docsrs, doc(cfg(not(windows))))]
     pub fn control_directory(&mut self, p: impl AsRef<Path>) -> &mut Self {
         self.control_dir = Some(p.as_ref().to_path_buf());
         self
@@ -105,7 +144,35 @@ impl SessionBuilder {
         self
     }
 
-    /// Connect to the host at the given `host` over SSH.
+    /// Enable or disable compression (including stdin, stdout, stderr, data
+    /// for forwarded TCP and unix-domain connections, sftp and scp
+    /// connections).
+    ///
+    /// Note that the ssh server can forcibly disable the compression.
+    ///
+    /// By default, ssh uses configure value set in `~/.ssh/config`.
+    ///
+    /// If `~/.ssh/config` does not enable compression, then it is disabled
+    /// by default.
+    pub fn compression(&mut self, compression: bool) -> &mut Self {
+        self.compression = Some(compression);
+        self
+    }
+
+    /// Specify the path to the `known_hosts` file.
+    ///
+    /// The path provided may use tilde notation (`~`) to refer to the user's
+    /// home directory.
+    ///
+    /// The default is `~/.ssh/known_hosts` and `~/.ssh/known_hosts2`.
+    pub fn user_known_hosts_file(&mut self, user_known_hosts_file: impl AsRef<Path>) -> &mut Self {
+        self.user_known_hosts_file =
+            Some(user_known_hosts_file.as_ref().to_owned().into_boxed_path());
+        self
+    }
+
+    /// Connect to the host at the given `host` over SSH using process impl, which will
+    /// spawn a new ssh process for each `Child` created.
     ///
     /// The format of `destination` is the same as the `destination` argument to `ssh`. It may be
     /// specified as either `[user@]hostname` or a URI of the form `ssh://[user@]hostname[:port]`.
@@ -115,15 +182,40 @@ impl SessionBuilder {
     /// If connecting requires interactive authentication based on `STDIN` (such as reading a
     /// password), the connection will fail. Consider setting up keypair-based authentication
     /// instead.
+    #[cfg(feature = "process-mux")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "process-mux")))]
     pub async fn connect<S: AsRef<str>>(&self, destination: S) -> Result<Session, Error> {
         let destination = destination.as_ref();
         let (builder, destination) = self.resolve(destination);
-        builder.just_connect(destination).await
+        let tempdir = builder.launch_master(destination).await?;
+        Ok(process_impl::Session::new(tempdir, destination).into())
+    }
+
+    /// Connect to the host at the given `host` over SSH using native mux, which will
+    /// create a new local socket connection for each `Child` created.
+    ///
+    /// See the crate-level documentation for more details on the difference between native and process-based mux.
+    ///
+    /// The format of `destination` is the same as the `destination` argument to `ssh`. It may be
+    /// specified as either `[user@]hostname` or a URI of the form `ssh://[user@]hostname[:port]`.
+    /// A username or port that is specified in the connection string overrides the one set in the
+    /// builder (but does not change the builder).
+    ///
+    /// If connecting requires interactive authentication based on `STDIN` (such as reading a
+    /// password), the connection will fail. Consider setting up keypair-based authentication
+    /// instead.
+    #[cfg(feature = "native-mux")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "native-mux")))]
+    pub async fn connect_mux<S: AsRef<str>>(&self, destination: S) -> Result<Session, Error> {
+        let destination = destination.as_ref();
+        let (builder, destination) = self.resolve(destination);
+        let tempdir = builder.launch_master(destination).await?;
+        Ok(native_mux_impl::Session::new(tempdir).into())
     }
 
     fn resolve<'a, 'b>(&'a self, mut destination: &'b str) -> (Cow<'a, Self>, &'b str) {
-        // the "new" ssh://user@host:port form is not supported by all versions of ssh, so we
-        // always translate it into the option form.
+        // the "new" ssh://user@host:port form is not supported by all versions of ssh,
+        // so we always translate it into the option form.
         let mut user = None;
         let mut port = None;
         if destination.starts_with("ssh://") {
@@ -159,20 +251,27 @@ impl SessionBuilder {
         (Cow::Owned(with_overrides), destination)
     }
 
-    pub(crate) async fn just_connect<S: AsRef<str>>(&self, host: S) -> Result<Session, Error> {
-        let destination = host.as_ref();
+    async fn launch_master(&self, destination: &str) -> Result<TempDir, Error> {
+        let socketdir = if let Some(socketdir) = self.control_dir.as_ref() {
+            socketdir
+        } else {
+            get_default_control_dir()?
+        };
 
-        let defaultdir = Path::new("./").to_path_buf();
-        let socketdir = self.control_dir.as_ref().unwrap_or(&defaultdir);
         let dir = Builder::new()
             .prefix(".ssh-connection")
             .tempdir_in(socketdir)
             .map_err(Error::Master)?;
+
+        let log = dir.path().join("log");
+
         let mut init = process::Command::new("ssh");
 
         init.stdin(Stdio::null())
-            .stdout(Stdio::piped())
-            .stderr(Stdio::piped())
+            .stdout(Stdio::null())
+            .stderr(Stdio::null())
+            .arg("-E")
+            .arg(&log)
             .arg("-S")
             .arg(dir.path().join("master"))
             .arg("-M")
@@ -212,31 +311,30 @@ impl SessionBuilder {
             init.arg("-F").arg(config_file);
         }
 
-        init.arg(destination);
+        if let Some(compression) = self.compression {
+            let arg = if compression { "yes" } else { "no" };
 
-        // eprintln!("{:?}", init);
-
-        // we spawn and immediately wait, because the process is supposed to fork.
-        // note that we cannot use .output, since it _also_ tries to read all of stdout/stderr.
-        // if the call _didn't_ error, then the backgrounded ssh client will still hold onto those
-        // handles, and it's still running, so those reads will hang indefinitely.
-        let mut child = init.spawn().map_err(Error::Connect)?;
-        let stdout = child.stdout.take().unwrap();
-        let mut stderr = child.stderr.take().unwrap();
-        let status = child.wait().await.map_err(Error::Connect)?;
-
-        if !status.success() {
-            let mut err = String::new();
-            stderr.read_to_string(&mut err).await.unwrap();
-            return Err(Error::interpret_ssh_error(&err));
+            init.arg("-o").arg(format!("Compression={}", arg));
         }
 
-        Ok(Session {
-            ctl: dir,
-            addr: String::from(destination),
-            terminated: false,
-            master: std::sync::Mutex::new(Some((stdout, stderr))),
-        })
+        if let Some(user_known_hosts_file) = &self.user_known_hosts_file {
+            let mut option: OsString = "UserKnownHostsFile=".into();
+            option.push(&**user_known_hosts_file);
+            init.arg("-o").arg(option);
+        }
+
+        init.arg(destination);
+
+        // we spawn and immediately wait, because the process is supposed to fork.
+        let status = init.status().await.map_err(Error::Connect)?;
+
+        if !status.success() {
+            let output = fs::read_to_string(log).map_err(Error::Connect)?;
+
+            Err(Error::interpret_ssh_error(&output))
+        } else {
+            Ok(dir)
+        }
     }
 }
 
@@ -269,41 +367,46 @@ impl KnownHosts {
     }
 }
 
-#[test]
-fn resolve() {
-    let b = SessionBuilder::default();
-    let (b, d) = b.resolve("ssh://test-user@127.0.0.1:2222");
-    assert_eq!(b.port.as_deref(), Some("2222"));
-    assert_eq!(b.user.as_deref(), Some("test-user"));
-    assert_eq!(d, "127.0.0.1");
+#[cfg(test)]
+mod tests {
+    use super::SessionBuilder;
 
-    let b = SessionBuilder::default();
-    let (b, d) = b.resolve("ssh://test-user@opensshtest:2222");
-    assert_eq!(b.port.as_deref(), Some("2222"));
-    assert_eq!(b.user.as_deref(), Some("test-user"));
-    assert_eq!(d, "opensshtest");
+    #[test]
+    fn resolve() {
+        let b = SessionBuilder::default();
+        let (b, d) = b.resolve("ssh://test-user@127.0.0.1:2222");
+        assert_eq!(b.port.as_deref(), Some("2222"));
+        assert_eq!(b.user.as_deref(), Some("test-user"));
+        assert_eq!(d, "127.0.0.1");
 
-    let b = SessionBuilder::default();
-    let (b, d) = b.resolve("ssh://opensshtest:2222");
-    assert_eq!(b.port.as_deref(), Some("2222"));
-    assert_eq!(b.user.as_deref(), None);
-    assert_eq!(d, "opensshtest");
+        let b = SessionBuilder::default();
+        let (b, d) = b.resolve("ssh://test-user@opensshtest:2222");
+        assert_eq!(b.port.as_deref(), Some("2222"));
+        assert_eq!(b.user.as_deref(), Some("test-user"));
+        assert_eq!(d, "opensshtest");
 
-    let b = SessionBuilder::default();
-    let (b, d) = b.resolve("ssh://test-user@opensshtest");
-    assert_eq!(b.port.as_deref(), None);
-    assert_eq!(b.user.as_deref(), Some("test-user"));
-    assert_eq!(d, "opensshtest");
+        let b = SessionBuilder::default();
+        let (b, d) = b.resolve("ssh://opensshtest:2222");
+        assert_eq!(b.port.as_deref(), Some("2222"));
+        assert_eq!(b.user.as_deref(), None);
+        assert_eq!(d, "opensshtest");
 
-    let b = SessionBuilder::default();
-    let (b, d) = b.resolve("ssh://opensshtest");
-    assert_eq!(b.port.as_deref(), None);
-    assert_eq!(b.user.as_deref(), None);
-    assert_eq!(d, "opensshtest");
+        let b = SessionBuilder::default();
+        let (b, d) = b.resolve("ssh://test-user@opensshtest");
+        assert_eq!(b.port.as_deref(), None);
+        assert_eq!(b.user.as_deref(), Some("test-user"));
+        assert_eq!(d, "opensshtest");
 
-    let b = SessionBuilder::default();
-    let (b, d) = b.resolve("opensshtest");
-    assert_eq!(b.port.as_deref(), None);
-    assert_eq!(b.user.as_deref(), None);
-    assert_eq!(d, "opensshtest");
+        let b = SessionBuilder::default();
+        let (b, d) = b.resolve("ssh://opensshtest");
+        assert_eq!(b.port.as_deref(), None);
+        assert_eq!(b.user.as_deref(), None);
+        assert_eq!(d, "opensshtest");
+
+        let b = SessionBuilder::default();
+        let (b, d) = b.resolve("opensshtest");
+        assert_eq!(b.port.as_deref(), None);
+        assert_eq!(b.user.as_deref(), None);
+        assert_eq!(d, "opensshtest");
+    }
 }
