@@ -5,6 +5,10 @@ use crate::*;
 #[doc(hidden)]
 pub mod unreleased {}
 
+/// ## Added
+///  - [`Session::subsystem`]
+pub mod v0_9_1 {}
+
 /// No changes since 0.9.0-rc4.
 pub mod v0_9_0 {}
 
